#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
__all__ = ["FeatureFunction", "FeatureHolder"]

import typing

from rialto.maker.utils import feature_name
from rialto.metadata import FeatureMetadata, ValueType


class FeatureFunction:
    """
    A container for feature generating object

    contains a callable object, it's name, parameters
    """

    def __init__(self, name: str, callable_object: typing.Callable, value_type: ValueType = ValueType.nominal):
        self.name = name
        self.callable = callable_object
        self.parameters: typing.Dict[str, typing.Any] = {}
        self.dependencies: typing.List[str] = []
        self.type = value_type
        self.description = "basic feature"

    def __str__(self) -> str:
        """
        Serialize to string for logging

        :return: string serialized object
        """
        return (
            f"Name: {self.name}\n\t"
            f"Parameters: {self.parameters}\n\t"
            f"Type: {self.get_type()}\n\t"
            f"Description: {self.description}"
        )

    def metadata(self) -> FeatureMetadata:
        """
        Return functions metadata

        :return: metadata dict
        """
        return FeatureMetadata(name=self.get_feature_name(), value_type=self.type, description=self.description)

    def get_feature_name(self) -> str:
        """
        Get feature name enhanced by parameters

        :return: full feature name
        """
        return feature_name(self.name, self.parameters)

    def get_type(self) -> str:
        """
        Get feature value type

        :return: value type
        """
        return self.type.value


class FeatureHolder(list):
    """
    A container for FeatureFunctions used in the feature building process

    Basically just a plain list with unique type name
    """

    def __init__(self):
        super().__init__()

    def get_metadata(self) -> typing.List[FeatureMetadata]:
        """
        Concats metadata from all functions

        :return: List of metadata
        """
        return [func.metadata() for func in self]



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["Record"]

from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Optional

from pyspark.sql import Row
from pyspark.sql.types import (
    DateType,
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)


@dataclass
class Record:
    """Dataclass with information about one run of one pipeline."""

    job: str
    target: str
    date: datetime.date
    time: timedelta
    records: int
    status: str
    reason: str
    exception: Optional[str] = None
    run_timestamp: datetime.timestamp = datetime.now()

    def get_schema(self) -> StructType:
        """Retrieve schema of pyspark DataFrame"""
        return StructType(
            [
                StructField("job", StringType(), nullable=False),
                StructField("target", StringType(), nullable=False),
                StructField("date", DateType(), nullable=False),
                StructField("time", StringType(), nullable=False),
                StructField("records", IntegerType(), nullable=False),
                StructField("status", StringType(), nullable=False),
                StructField("reason", StringType(), nullable=False),
                StructField("exception", StringType(), nullable=True),
                StructField("run_timestamp", TimestampType(), nullable=False),
            ]
        )

    def to_spark_row(self) -> Row:
        """Convert Record to Spark Row"""
        return Row(
            job=self.job,
            target=self.target,
            date=self.date,
            time=str(self.time),
            records=self.records,
            status=self.status,
            reason=self.reason,
            exception=self.exception,
            run_timestamp=self.run_timestamp,
        )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = [
    "get_pipelines_config",
]

from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict

from rialto.common.utils import load_yaml
from rialto.runner.config_overrides import override_config


class BaseConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")


class IntervalConfig(BaseConfig):
    units: str
    value: int


class ScheduleConfig(BaseConfig):
    frequency: str
    day: Optional[int] = 0
    info_date_shift: Optional[List[IntervalConfig]] = IntervalConfig(units="days", value=0)


class DependencyConfig(BaseConfig):
    table: str
    name: Optional[str] = None
    date_col: str
    interval: IntervalConfig


class ModuleConfig(BaseConfig):
    python_module: str
    python_class: str


class MailConfig(BaseConfig):
    sender: str
    to: List[str]
    smtp: str
    subject: str
    sent_empty: Optional[bool] = False


class RunnerConfig(BaseConfig):
    watched_period_units: str
    watched_period_value: int
    mail: Optional[MailConfig] = None
    bookkeeping: Optional[str] = None


class TargetConfig(BaseConfig):
    target_schema: str
    target_partition_column: str


class MetadataManagerConfig(BaseConfig):
    metadata_schema: str


class FeatureLoaderConfig(BaseConfig):
    feature_schema: str
    metadata_schema: str


class PipelineConfig(BaseConfig):
    name: str
    module: ModuleConfig
    schedule: ScheduleConfig
    dependencies: Optional[List[DependencyConfig]] = []
    target: TargetConfig = None
    metadata_manager: Optional[MetadataManagerConfig] = None
    feature_loader: Optional[FeatureLoaderConfig] = None
    extras: Optional[Dict] = {}


class PipelinesConfig(BaseConfig):
    runner: RunnerConfig
    pipelines: list[PipelineConfig]


def get_pipelines_config(path: str, overrides: Dict) -> PipelinesConfig:
    """Load and parse yaml config"""
    raw_config = load_yaml(path)
    if overrides:
        cfg = override_config(raw_config, overrides)
        return PipelinesConfig(**cfg)
    else:
        return PipelinesConfig(**raw_config)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["class_to_catalog_name"]


def class_to_catalog_name(class_name) -> str:
    """
    Map python class name of feature group (CammelCase) to databricks compatible format (lowercase with underscores)

    :param class_name: Python class name
    :return: feature storage name
    """
    res = []
    for i in range(0, len(class_name)):
        c = class_name[i]
        if c.isupper():
            if i != 0:
                res.append("_")
            res.append(c.lower())
        else:
            res.append(c)
    return "".join(res)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["Table"]

from rialto.metadata import class_to_catalog_name


class Table:
    """Handler for databricks catalog paths"""

    def __init__(
        self,
        catalog: str = None,
        schema: str = None,
        table: str = None,
        schema_path: str = None,
        table_path: str = None,
        class_name: str = None,
        partition: str = None,
    ):
        self.catalog = catalog
        self.schema = schema
        self.table = table
        self.partition = partition
        if schema_path:
            schema_path = schema_path.split(".")
            self.catalog = schema_path[0]
            self.schema = schema_path[1]
        if table_path:
            table_path = table_path.split(".")
            self.catalog = table_path[0]
            self.schema = table_path[1]
            self.table = table_path[2]
        if class_name:
            self.table = class_to_catalog_name(class_name)

    def get_schema_path(self):
        """Get path of table's schema"""
        return f"{self.catalog}.{self.schema}"

    def get_table_path(self):
        """Get full table path"""
        return f"{self.catalog}.{self.schema}.{self.table}"



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["load_module", "table_exists", "get_partitions", "init_tools", "find_dependency"]

from datetime import date
from importlib import import_module
from typing import List, Tuple

from pyspark.sql import SparkSession

from rialto.common import DataReader
from rialto.loader import PysparkFeatureLoader
from rialto.metadata import MetadataManager
from rialto.runner.config_loader import ModuleConfig, PipelineConfig
from rialto.runner.table import Table
from rialto.runner.transformation import Transformation


def load_module(cfg: ModuleConfig) -> Transformation:
    """
    Load feature group

    :param cfg: Feature configuration
    :return: Transformation object
    """
    module = import_module(cfg.python_module)
    class_obj = getattr(module, cfg.python_class)
    return class_obj()


def table_exists(spark: SparkSession, table: str) -> bool:
    """
    Check table exists in spark catalog

    :param table: full table path
    :return: bool
    """
    return spark.catalog.tableExists(table)


def get_partitions(reader: DataReader, table: Table) -> List[date]:
    """
    Get partition values

    :param table: Table object
    :return: List of partition values
    """
    rows = (
        reader.get_table(table.get_table_path(), date_column=table.partition)
        .select(table.partition)
        .distinct()
        .collect()
    )
    return [r[table.partition] for r in rows]


def init_tools(spark: SparkSession, pipeline: PipelineConfig) -> Tuple[MetadataManager, PysparkFeatureLoader]:
    """
    Initialize metadata manager and feature loader

    :param spark: Spark session
    :param pipeline: Pipeline configuration
    :return: MetadataManager and PysparkFeatureLoader
    """
    if pipeline.metadata_manager is not None:
        metadata_manager = MetadataManager(spark, pipeline.metadata_manager.metadata_schema)
    else:
        metadata_manager = None

    if pipeline.feature_loader is not None:
        feature_loader = PysparkFeatureLoader(
            spark,
            feature_schema=pipeline.feature_loader.feature_schema,
            metadata_schema=pipeline.feature_loader.metadata_schema,
        )
    else:
        feature_loader = None
    return metadata_manager, feature_loader


def find_dependency(config: PipelineConfig, name: str):
    """
    Get dependency from config

    :param config: Pipeline configuration
    :param name: Dependency name
    :return: Dependency object
    """
    for dep in config.dependencies:
        if dep.name == name:
            return dep
    return None



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["GroupMetadata"]

from dataclasses import dataclass
from typing import List, Tuple

from pyspark.sql import Row
from typing_extensions import Self

from rialto.metadata.enums import Schedule
from rialto.metadata.utils import class_to_catalog_name


@dataclass
class GroupMetadata:
    """A dataclass to hold all information about a feature group"""

    name: str
    frequency: Schedule
    description: str
    key: List[str]
    owner: str
    fs_name: str = None
    features: List[str] = None

    def __repr__(self) -> str:
        """Serialize object to string"""
        return (
            "GroupMetadata("
            f"name={self.name!r}, frequency={self.frequency!r}, "
            f"feature store name={self.fs_name!r},"
            f"description={self.description!r}, key={self.key!r}, "
            f"features={self.features!r}, owner={self.owner!r}"
            ")"
        )

    def add_features(self, features: List[str]) -> Self:
        """
        Add feature list belonging to the group

        :param group: list of feature names
        :return: self
        """
        if len(features):
            self.features = features
        return self

    def to_tuple(self) -> Tuple:
        """
        Serialize to tuple

        :return: tuple with feature group information
        """
        if not self.fs_name:
            self.fs_name = class_to_catalog_name(self.name)
        return (self.name, self.frequency.value, self.description, self.key, self.fs_name, self.owner)

    @classmethod
    def from_spark(cls, schema: Row) -> Self:
        """
        Create new instance from spark row

        :param record: spark row
        :return: new instance
        """
        return GroupMetadata(
            name=schema.group_name,
            fs_name=schema.group_fs_name,
            frequency=Schedule[schema.group_frequency],
            description=schema.group_description,
            key=schema.group_key,
            owner=schema.group_owner
        )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["get_feature_config"]

from typing import List, Optional

from pydantic import BaseModel

from rialto.common.utils import load_yaml


class GroupConfig(BaseModel):
    group: str
    prefix: str
    features: List[str]


class BaseConfig(BaseModel):
    group: str
    keys: List[str]


class FeatureConfig(BaseModel):
    selection: List[GroupConfig]
    base: BaseConfig
    maps: Optional[List[str]] = None


def get_feature_config(path) -> FeatureConfig:
    """
    Read yaml and parse it

    :param path: config path
    :return: Pydantic feature config object
    """
    return FeatureConfig(**load_yaml(path))



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["feature_name"]

import typing


def _get_feature_parameter_suffix(parameters: typing.Dict) -> str:
    """
    Collate all parameter names and their value in format p1_v1_p2_v2...

    :param parameters: unordered dict of parameters and values
    :return: string serialized parameters
    """
    if len(parameters) == 0:
        return ""
    params = ""
    for parameter_name, value in sorted(parameters.items()):
        params += f"_{parameter_name}_{value}"
    return params


def feature_name(name: str, parameters: typing.Dict) -> str:
    """
    Join feature function name with parameters to create a feature name

    :param name: string name of feature
    :param parameters: unordered dict of parameters and values
    :return: feature name
    """
    params = _get_feature_parameter_suffix(parameters)
    if len(params) == 0:
        return name.upper()
    return f"{name}{params}".upper()



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["JobBase"]

import abc
import datetime
import typing

import pyspark.sql.functions as F
from loguru import logger
from pydantic import BaseModel
from pyspark.sql import DataFrame, SparkSession

from rialto.common import TableReader
from rialto.jobs.resolver import Resolver
from rialto.loader import PysparkFeatureLoader
from rialto.metadata import MetadataManager
from rialto.runner import Transformation
from rialto.runner.config_loader import PipelineConfig


class JobMetadata(BaseModel):
    job_name: str
    dist_name: str
    dist_version: str


class JobBase(Transformation):
    """A Base Class for Rialto Jobs. Serves as a foundation into which the @job decorators are inserted."""

    @abc.abstractmethod
    def get_custom_callable(self) -> typing.Callable:
        """Getter - Custom callable (i.e. job transformation function)"""
        pass

    @abc.abstractmethod
    def get_job_metadata(self) -> JobMetadata:
        """Job metadata getter"""
        pass

    @abc.abstractmethod
    def get_job_name(self) -> str:
        """Job name getter"""
        pass

    @abc.abstractmethod
    def get_disable_version(self) -> bool:
        """Disable version getter"""
        pass

    def _get_resolver(
        self,
        spark: SparkSession,
        run_date: datetime.date,
        table_reader: TableReader,
        config: PipelineConfig = None,
        metadata_manager: MetadataManager = None,
        feature_loader: PysparkFeatureLoader = None,
    ) -> Resolver:
        resolver = Resolver()

        # Static Always - Available dependencies
        resolver.register_object(spark, "spark")
        resolver.register_object(run_date, "run_date")
        resolver.register_object(config, "config")
        resolver.register_object(self.get_job_metadata(), "job_metadata")
        resolver.register_object(table_reader, "table_reader")

        # Optionals
        if feature_loader is not None:
            resolver.register_object(feature_loader, "feature_loader")

        if metadata_manager is not None:
            resolver.register_object(metadata_manager, "metadata_manager")

        return resolver

    def _get_timestamp_holder_result(self, spark) -> DataFrame:
        return spark.createDataFrame(
            [(self.get_job_name(), datetime.datetime.now())], schema="JOB_NAME string, CREATION_TIME timestamp"
        )

    def _add_job_version(self, df: DataFrame) -> DataFrame:
        version = self.get_job_metadata().dist_version
        return df.withColumn("VERSION", F.lit(version))

    def run(
        self,
        reader: TableReader,
        run_date: datetime.date,
        spark: SparkSession = None,
        config: PipelineConfig = None,
        metadata_manager: MetadataManager = None,
        feature_loader: PysparkFeatureLoader = None,
    ) -> DataFrame:
        """
        Rialto transformation run

        :param reader: data store api object
        :param info_date: date
        :param spark: spark session
        :param config: pipeline config
        :return: dataframe
        """
        try:
            resolver = self._get_resolver(spark, run_date, reader, config, metadata_manager, feature_loader)

            custom_callable = self.get_custom_callable()
            raw_result = resolver.resolve(custom_callable)

            if raw_result is None:
                raw_result = self._get_timestamp_holder_result(spark)

            if not self.get_disable_version():
                return self._add_job_version(raw_result)

            return raw_result

        except Exception as e:
            logger.exception(e)
            raise e



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["DataReader", "TableReader"]

import abc
import datetime
from typing import Optional

import pyspark.sql.functions as F
from pyspark.sql import DataFrame, SparkSession


class DataReader(metaclass=abc.ABCMeta):
    """
    This is an abstract class defining interface for reader of spark tables

    Data reader provides to public functions, get_latest and get_table.
    get_latest reads a single snapshot of the given table, while get_table reads the whole table or multiple snapshots.
    """

    @abc.abstractmethod
    def get_latest(
        self,
        table: str,
        date_column: str,
        date_until: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """
        Get latest available date partition of the table until specified date

        :param table: input table path
        :param date_until: Optional until date (inclusive)
        :param uppercase_columns: Option to refactor all column names to uppercase
        :return: Dataframe
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get_table(
        self,
        table: str,
        date_column: str,
        date_from: Optional[datetime.date] = None,
        date_to: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """
        Get a whole table or a slice by selected dates

        :param table: input table path
        :param date_from: Optional date from (inclusive)
        :param date_to: Optional date to (inclusive)
        :param uppercase_columns: Option to refactor all column names to uppercase
        :return: Dataframe
        """
        raise NotImplementedError


class TableReader(DataReader):
    """An implementation of data reader for databricks tables"""

    def __init__(self, spark: SparkSession):
        """
        Init

        :param spark:
        """
        self.spark = spark
        super().__init__()

    def _uppercase_column_names(self, df: DataFrame) -> DataFrame:
        """
        Change the case of all column names to uppercase

        :param df: Dataframe
        :return: renamed Dataframe
        """
        for col in df.columns:
            df = df.withColumnRenamed(col, col.upper())
        return df

    def _get_latest_available_date(self, df: DataFrame, date_col: str, until: Optional[datetime.date]) -> datetime.date:
        if until:
            df = df.filter(F.col(date_col) <= until)
        df = df.select(F.max(date_col)).alias("latest")
        return df.head()[0]

    def get_latest(
        self,
        table: str,
        date_column: str,
        date_until: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """
        Get latest available date partition of the table until specified date

        :param table: input table path
        :param date_until: Optional until date (inclusive)
        :param date_column: column to filter dates on, takes highest priority
        :param uppercase_columns: Option to refactor all column names to uppercase
        :return: Dataframe
        """
        df = self.spark.read.table(table)

        selected_date = self._get_latest_available_date(df, date_column, date_until)
        df = df.filter(F.col(date_column) == selected_date)

        if uppercase_columns:
            df = self._uppercase_column_names(df)
        return df

    def get_table(
        self,
        table: str,
        date_column: str,
        date_from: Optional[datetime.date] = None,
        date_to: Optional[datetime.date] = None,
        uppercase_columns: bool = False,
    ) -> DataFrame:
        """
        Get a whole table or a slice by selected dates

        :param table: input table path
        :param date_from: Optional date from (inclusive)
        :param date_to: Optional date to (inclusive)
        :param date_column: column to filter dates on, takes highest priority
        :param uppercase_columns: Option to refactor all column names to uppercase
        :return: Dataframe
        """
        df = self.spark.read.table(table)

        if date_from:
            df = df.filter(F.col(date_column) >= date_from)
        if date_to:
            df = df.filter(F.col(date_column) <= date_to)
        if uppercase_columns:
            df = self._uppercase_column_names(df)
        return df



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["ResolverException", "Resolver"]

import inspect
import typing

from rialto.jobs.module_register import ModuleRegister


class ResolverException(Exception):
    """Resolver Errors Class - In Most Cases your dependency tree is not complete."""

    pass


class Resolver:
    """
    Resolver handles dependency management between datasets and jobs.

    We register different callables, which can depend on other callables.
    Calling resolve() we attempt to resolve these dependencies.
    """

    def __init__(self):
        self._storage = {}

    def register_object(self, object: typing.Any, name: str) -> None:
        """
        Register an object with a given name for later resolution.

        :param object: object to register (getter)
        :param name: str, custom name
        :return: None
        """

        self.register_getter(lambda: object, name)

    def register_getter(self, callable: typing.Callable, name: str = None) -> str:
        """
        Register callable with a given name for later resolution.

        In case name isn't present, function's __name__ attribute will be used.

        :param callable: callable to register (getter)
        :param name: str, custom name, f.__name__ will be used otherwise
        :return: str, name under which the callable has been registered
        """
        if name is None:
            name = getattr(callable, "__name__", repr(callable))

        if name in self._storage:
            raise ResolverException(f"Resolver already registered {name}!")

        self._storage[name] = callable
        return name

    def _find_getter(self, name: str, module_name) -> typing.Callable:
        if name in self._storage.keys():
            return self._storage[name]

        callable_from_dependencies = ModuleRegister.find_callable(name, module_name)
        if callable_from_dependencies is None:
            raise ResolverException(f"{name} declaration not found!")

        return callable_from_dependencies

    def resolve(self, callable: typing.Callable) -> typing.Dict[str, typing.Any]:
        """
        Take a callable and resolve its dependencies / arguments. Arguments can be
        a) objects registered via register_object
        b) callables registered via register_getter
        c) ModuleRegister registered callables via ModuleRegister.register_callable (+ dependencies)

        Arguments are resolved recursively according to requirements; For example, if we have
        a(b, c), b(d), and c(), d() registered, then we recursively call resolve() methods until we resolve
        c, d -> b -> a

        :param callable: function to resolve
        :return: result of the callable
        """

        arg_list = {}

        signature = inspect.signature(callable)
        module_name = callable.__module__

        for param in signature.parameters.values():
            param_getter = self._find_getter(param.name, module_name)
            arg_list[param.name] = self.resolve(param_getter)

        return callable(**arg_list)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["Transformation"]

import abc
import datetime

from pyspark.sql import DataFrame, SparkSession

from rialto.common import DataReader
from rialto.loader import PysparkFeatureLoader
from rialto.metadata import MetadataManager
from rialto.runner.config_loader import PipelineConfig


class Transformation(metaclass=abc.ABCMeta):
    """Interface for feature implementation"""

    @abc.abstractmethod
    def run(
        self,
        reader: DataReader,
        run_date: datetime.date,
        spark: SparkSession = None,
        config: PipelineConfig = None,
        metadata_manager: MetadataManager = None,
        feature_loader: PysparkFeatureLoader = None,
    ) -> DataFrame:
        """
        Run the transformation

        :param reader: data store api object
        :param run_date: date
        :param spark: spark session
        :param config: pipeline config
        :param metadata_manager: metadata manager
        :param feature_loader: feature loader
        :return: dataframe
        """
        raise NotImplementedError



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["FeatureMaker"]

import datetime
import inspect
import types
import typing

import pyspark.sql.functions as F
from loguru import logger
from pyspark.sql import DataFrame

from rialto.maker.containers import FeatureFunction, FeatureHolder


class _FeatureMaker:
    """
    A framework for feature making

    Enables registration of callable feature functions and then executes them.
    """

    def __init__(self):
        self.feature_functions = FeatureHolder()
        self.data_frame = None
        self.make_date = None
        self.key = None

    def _set_values(self, df: DataFrame, key: typing.Union[str, typing.List[str]], make_date: datetime.date):
        """
        Instance value setter

        :param df: DataFrame with input data
        :param key: simple or compound string key
        :param make_date: a date to make feature from
        :return: None
        """
        self.data_frame = df
        if isinstance(key, str):
            self.key = [key]
        else:
            self.key = key
        self.make_date = make_date

    def _order_by_dependencies(self, feature_holders: typing.List[FeatureHolder]) -> typing.List[FeatureHolder]:
        """
        Order features like directional graph

        Simple O(n^2) solution, in each pass try to find one feature group with dependencies resolved
        :param feature_holders: List of feature holders each with one group of feature_functions
        :return: ordered list of feature holders each with one group of feature_functions
        """
        logger.trace("Resolving module dependencies")

        ordered = []
        resolved = set()
        leftover = feature_holders
        while True:
            dep_is_resolved = True
            i = 0
            for i in range(len(leftover)):
                dep_is_resolved = True
                for dep in leftover[i][0].dependencies:
                    if dep not in resolved:
                        dep_is_resolved = False
                        break
                if dep_is_resolved:
                    break
            if dep_is_resolved:
                resolved.add(leftover[i][0].name)
                ordered.append(leftover.pop(i))
            else:
                raise Exception("Feature dependencies can't be resolved!")

            if len(leftover) == 0:
                break
        return ordered

    def _load_features(self, features_module: types.ModuleType) -> typing.List:
        """
        Find feature function definitions inside a given python module

        :param features_module: a python module with definitions of feature functions using wrappers
        :return: List of FeatureHolders
        """
        return [value for (member, value) in inspect.getmembers(features_module) if isinstance(value, FeatureHolder)]

    def _register_module(self, features_module: types.ModuleType) -> FeatureHolder:
        """
        Find feature function definitions inside a given python module and registers them for use in make call

        :param features_module: a python module with definitions of feature functions using FML wrappers
        :return: FeatureHolder
        """
        logger.trace(f"Registering module: {features_module}")
        feature_functions = FeatureHolder()

        ordered_feature_holders = self._order_by_dependencies(self._load_features(features_module))
        for feature_holder in ordered_feature_holders:
            for feature_function in feature_holder:
                logger.trace(f"Registering feature:\n\t{feature_function}")
                feature_functions.append(feature_function)
        return feature_functions

    def _filter_null_keys(self, df: DataFrame):
        if isinstance(self.key, str):
            return df.filter(F.col(self.key).isNotNull())
        else:
            for k in self.key:
                df = df.filter(F.col(k).isNotNull())
        return df

    def _make_sequential(self, keep_preexisting: bool) -> DataFrame:
        """
        Make features by creating new columns on the existing dataframe

        :param keep_preexisting: Bool, keep preexisting data in the dataframe after making features
        :return: DataFrame
        """
        feature_names = []
        for feature_function in self.feature_functions:
            logger.trace(f"Making feature:\n\t{feature_function}")
            feature_names.append(feature_function.get_feature_name())
            self.data_frame = self.data_frame.withColumn(
                feature_function.get_feature_name(), feature_function.callable()
            )
        if not keep_preexisting:
            logger.info("Dropping non-selected columns")
            self.data_frame = self.data_frame.select(*self.key, *feature_names)
        return self._filter_null_keys(self.data_frame)

    def _make_aggregated(self) -> DataFrame:
        """
        Make features by creating new columns via aggregation on defined key

        :return: DataFrame
        """
        aggregates = []
        for feature_function in self.feature_functions:
            logger.trace(f"Creating aggregate: \n\t{feature_function.__str__()}")
            aggregates.append(feature_function.callable().alias(feature_function.get_feature_name()))

        self.data_frame = self.data_frame.groupBy(self.key).agg(*aggregates)
        return self._filter_null_keys(self.data_frame)

    def make(
        self,
        df: DataFrame,
        key: typing.Union[str, typing.List[str]],
        make_date: datetime.date,
        features_module: types.ModuleType,
        keep_preexisting: bool = False,
    ) -> (DataFrame, typing.Dict):
        """
        Make features by creating new columns based on definitions in imported module

        :param df: DataFrame with input data
        :param key: simple or compound string key
        :param make_date: a date to make feature from
        :param features_module: a python module with definitions of feature functions using FML wrappers
        :param keep_preexisting: bool to decide whether to keep input data
        :return: DataFrame with features, Dict with feature metadata
        """
        logger.info(f"Making sequential for \n\tkey: {key} \n\ton {make_date} \n\tfrom {features_module}")
        self._set_values(df, key, make_date)
        self.feature_functions = self._register_module(features_module)
        features = self._make_sequential(keep_preexisting)
        logger.info(f"Finished making sequential features from {features_module}")
        return features, self.feature_functions.get_metadata()

    def make_aggregated(
        self,
        df: DataFrame,
        key: typing.Union[str, typing.List[str]],
        make_date: datetime.date,
        features_module: types.ModuleType,
    ) -> (DataFrame, typing.Dict):
        """
        Make features by creating new columns based on definitions in imported module

        :param df: DataFrame with input data
        :param key: simple or compound string key
        :param make_date: a date to make feature from
        :param features_module: a python module with definitions of feature functions using FML wrappers
        :return: DataFrame with features, Dict with feature metadata
        """
        logger.info(f"Making aggregated for \n\tkey: {key} \n\ton {make_date} \n\tfrom {features_module}")
        self._set_values(df, key, make_date)
        self.feature_functions = self._register_module(features_module)
        features = self._make_aggregated()
        logger.info(f"Finished making aggregated features from {features_module}")
        return features, self.feature_functions.get_metadata()

    def _find_feature(self, name: str, feature_functions: FeatureHolder) -> FeatureFunction:
        """
        Find a single feature function by name

        :param name: str name of the feature
        :param feature_functions: FeatureHolder with all feature functions
        :return: wanted FeatureFunction
        """
        for feature_function in feature_functions:
            if feature_function.get_feature_name() == name:
                return feature_function
        raise Exception(f"Feature {name} is not defined!")

    def make_single_feature(
        self,
        df: DataFrame,
        name: typing.Union[str, typing.List[str]],
        features_module: types.ModuleType,
        make_date: datetime.date = None,
    ) -> DataFrame:
        """
        Make single feature by creating new columns based on definition in imported module

        Intended for being able to test single features
        :param df: DataFrame with input data
        :param name: name of the feature
        :param features_module: a python module with definitions of feature functions using FML wrappers
        :param make_date: an optional date to make feature from
        :return: DataFrame with features
        """
        self.make_date = make_date
        feature_functions = self._register_module(features_module)
        feature = self._find_feature(name, feature_functions)
        return df.withColumn(feature.get_feature_name(), feature.callable()).select(feature.get_feature_name())

    def make_single_agg_feature(
        self,
        df: DataFrame,
        name: str,
        key: typing.Union[str, typing.List[str]],
        features_module: types.ModuleType,
        make_date: datetime.date = None,
    ) -> DataFrame:
        """
        Make single feature by creating new columns based on definition in imported module

        Intended for being able to test single features
        :param df: DataFrame with input data
        :param name: name of the feature
        :param key: key to aggregate on
        :param features_module: a python module with definitions of feature functions using FML wrappers
        :param make_date: an optional date to make feature from
        :return: DataFrame with features
        """
        self.make_date = make_date
        feature_functions = self._register_module(features_module)
        feature = self._find_feature(name, feature_functions)
        return df.groupBy(key).agg(feature.callable().alias(feature.get_feature_name()))


FeatureMaker = _FeatureMaker()



import os
import re

import yaml
from loguru import logger

__all__ = ["EnvLoader"]

# Regex pattern to capture variable and the rest of the string
_path_matcher = re.compile(r"(?P<before>.*)\$\{(?P<env_name>[^}^{:]+)(?::(?P<default_value>[^}^{]*))?\}(?P<after>.*)")


def _path_constructor(loader, node):
    value = node.value
    match = _path_matcher.search(value)
    if match:
        before = match.group("before")
        after = match.group("after")
        sub = os.getenv(match.group("env_name"), match.group("default_value"))
        if sub is None:
            raise ValueError(f"Environment variable {match.group('env_name')} has no assigned value")
        new_value = before + sub + after
        logger.info(f"Config: Replacing {value}, with {new_value}")
        return new_value
    return value


class EnvLoader(yaml.SafeLoader):
    """Custom loader that replaces values with environment variables"""

    pass


EnvLoader.add_implicit_resolver("!env_substitute", _path_matcher, None)
EnvLoader.add_constructor("!env_substitute", _path_constructor)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["DateManager"]

from datetime import date, datetime
from typing import List

from dateutil.relativedelta import relativedelta

from rialto.runner.config_loader import ScheduleConfig


class DateManager:
    """Date generation and shifts based on configuration"""

    @staticmethod
    def str_to_date(str_date: str) -> date:
        """
        Convert YYYY-MM-DD string to date

        :param str_date: string date
        :return: date
        """
        return datetime.strptime(str_date, "%Y-%m-%d").date()

    @staticmethod
    def date_subtract(run_date: date, units: str, value: int) -> date:
        """
        Generate starting date from given date and config

        :param run_date: base date
        :param units: units: years, months, weeks, days
        :param value: number of units to subtract
        :return: Starting date
        """
        if units == "years":
            return run_date - relativedelta(years=value)
        if units == "months":
            return run_date - relativedelta(months=value)
        if units == "weeks":
            return run_date - relativedelta(weeks=value)
        if units == "days":
            return run_date - relativedelta(days=value)
        raise ValueError(f"Unknown time unit {units}")

    @staticmethod
    def all_dates(date_from: date, date_to: date) -> List[date]:
        """
        Get list of all dates between, inclusive

        :param date_from: starting date
        :param date_to: ending date
        :return: List[date]
        """
        if date_to < date_from:
            date_to, date_from = date_from, date_to

        return [date_from + relativedelta(days=n) for n in range((date_to - date_from).days + 1)]

    @staticmethod
    def run_dates(date_from: date, date_to: date, schedule: ScheduleConfig) -> List[date]:
        """
        Select dates inside given interval depending on frequency and selected day

        :param date_from: interval start
        :param date_to: interval end
        :param schedule: schedule config
        :return: list of dates
        """
        options = DateManager.all_dates(date_from, date_to)
        if schedule.frequency == "daily":
            return options
        if schedule.frequency == "weekly":
            return [x for x in options if x.isoweekday() == schedule.day]
        if schedule.frequency == "monthly":
            return [x for x in options if x.day == schedule.day]
        raise ValueError(f"Unknown frequency {schedule.frequency}")

    @staticmethod
    def to_info_date(date: date, schedule: ScheduleConfig) -> date:
        """
        Shift given date according to config

        :param date: input date
        :param schedule: schedule config
        :return: date
        """
        if isinstance(schedule.info_date_shift, List):
            for shift in schedule.info_date_shift:
                date = DateManager.date_subtract(date, units=shift.units, value=shift.value)
        else:
            date = DateManager.date_subtract(
                date, units=schedule.info_date_shift.units, value=schedule.info_date_shift.value
            )
        return date



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["Runner"]

import datetime
from datetime import date
from typing import Dict, List, Tuple

import pyspark.sql.functions as F
from loguru import logger
from pyspark.sql import DataFrame, SparkSession

import rialto.runner.utils as utils
from rialto.common import TableReader
from rialto.runner.config_loader import PipelineConfig, get_pipelines_config
from rialto.runner.date_manager import DateManager
from rialto.runner.reporting.record import Record
from rialto.runner.reporting.tracker import Tracker
from rialto.runner.table import Table
from rialto.runner.transformation import Transformation


class Runner:
    """A scheduler and dependency checker for feature runs"""

    def __init__(
        self,
        spark: SparkSession,
        config_path: str,
        run_date: str = None,
        rerun: bool = False,
        op: str = None,
        skip_dependencies: bool = False,
        overrides: Dict = None,
    ):
        self.spark = spark
        self.config = get_pipelines_config(config_path, overrides)
        self.reader = TableReader(spark)
        self.rerun = rerun
        self.skip_dependencies = skip_dependencies
        self.op = op
        self.tracker = Tracker(
            mail_cfg=self.config.runner.mail, bookkeeping=self.config.runner.bookkeeping, spark=spark
        )

        if run_date:
            run_date = DateManager.str_to_date(run_date)
        else:
            run_date = date.today()

        self.date_from = DateManager.date_subtract(
            run_date=run_date,
            units=self.config.runner.watched_period_units,
            value=self.config.runner.watched_period_value,
        )

        self.date_until = run_date

        if self.date_from > self.date_until:
            raise ValueError(f"Invalid date range from {self.date_from} until {self.date_until}")
        logger.info(f"Running period set to: {self.date_from} - {self.date_until}")

    def _execute(self, instance: Transformation, run_date: date, pipeline: PipelineConfig) -> DataFrame:
        """
        Run the job

        :param instance: Instance of Transformation
        :param run_date: date to run for
        :param pipeline: pipeline configuration
        :return: Dataframe
        """
        metadata_manager, feature_loader = utils.init_tools(self.spark, pipeline)

        df = instance.run(
            spark=self.spark,
            run_date=run_date,
            config=pipeline,
            reader=self.reader,
            metadata_manager=metadata_manager,
            feature_loader=feature_loader,
        )

        return df

    def _write(self, df: DataFrame, info_date: date, table: Table) -> None:
        """
        Write dataframe to storage

        :param df: dataframe to write
        :param info_date: date to partition
        :param table: path to write to
        :return: None
        """
        df = df.withColumn(table.partition, F.lit(info_date))
        df.write.partitionBy(table.partition).mode("overwrite").saveAsTable(table.get_table_path())
        logger.info(f"Results writen to {table.get_table_path()}")

    def _check_written(self, info_date: date, table: Table) -> int:
        """
        Check if there are records written for given date

        :param info_date: date to check
        :param table: target table object
        :return: number of records
        """
        df = self.spark.read.table(table.get_table_path())
        df = df.filter(F.col(table.partition) == info_date)
        return df.count()

    def check_dates_have_partition(self, table: Table, dates: List[date]) -> List[bool]:
        """
        For given list of dates, check if there is a matching partition for each

        :param table: Table object
        :param dates: list of dates to check
        :return: list of bool
        """
        if utils.table_exists(self.spark, table.get_table_path()):
            partitions = utils.get_partitions(self.reader, table)
            return [(date in partitions) for date in dates]
        else:
            logger.info(f"Table {table.get_table_path()} doesn't exist!")
            return [False for _ in dates]

    def check_dependencies(self, pipeline: PipelineConfig, run_date: date) -> bool:
        """
        Check for all dependencies in config if they have available partitions

        :param pipeline: configuration
        :param run_date: run date
        :return: bool
        """
        logger.info(f"{pipeline.name} checking dependencies for {run_date}")

        error = ""

        for dependency in pipeline.dependencies:
            dep_from = DateManager.date_subtract(run_date, dependency.interval.units, dependency.interval.value)
            logger.info(f"Looking for {dependency.table} from {dep_from} until {run_date}")

            possible_dep_dates = DateManager.all_dates(dep_from, run_date)

            logger.debug(f"Date column for {dependency.table} is {dependency.date_col}")

            source = Table(table_path=dependency.table, partition=dependency.date_col)
            if True in self.check_dates_have_partition(source, possible_dep_dates):
                logger.info(f"Dependency for {dependency.table} from {dep_from} until {run_date} is fulfilled")
            else:
                msg = f"Missing dependency for {dependency.table} from {dep_from} until {run_date}"
                logger.info(msg)
                error = error + msg + "\n"

        if error != "":
            self.tracker.last_error = error
            return False

        return True

    def _get_completion(self, target: Table, info_dates: List[date]) -> List[bool]:
        """
        Check if model has run for given dates

        :param target_path: Table object
        :param info_dates: list of dates
        :return: bool list
        """
        if self.rerun:
            return [False for _ in info_dates]
        else:
            return self.check_dates_have_partition(target, info_dates)

    def _select_run_dates(self, pipeline: PipelineConfig, table: Table) -> Tuple[List, List]:
        """
        Select run dates and info dates based on completion

        :param pipeline: pipeline config
        :param table: table path
        :return: list of run dates and list of info dates
        """
        possible_run_dates = DateManager.run_dates(self.date_from, self.date_until, pipeline.schedule)
        possible_info_dates = [DateManager.to_info_date(x, pipeline.schedule) for x in possible_run_dates]
        current_state = self._get_completion(table, possible_info_dates)

        selection = [
            (run, info) for run, info, state in zip(possible_run_dates, possible_info_dates, current_state) if not state
        ]

        if not len(selection):
            logger.info(f"{pipeline.name} has no dates to run")
            return [], []

        selected_run_dates, selected_info_dates = zip(*selection)
        logger.info(f"{pipeline.name} identified to run for {selected_run_dates}")

        return list(selected_run_dates), list(selected_info_dates)

    def _run_one_date(self, pipeline: PipelineConfig, run_date: date, info_date: date, target: Table) -> int:
        """
        Run one pipeline for one date

        :param pipeline: pipeline cfg
        :param run_date: run date
        :param info_date: information date
        :param target: target Table
        :return: success bool
        """
        if self.skip_dependencies or self.check_dependencies(pipeline, run_date):
            logger.info(f"Running {pipeline.name} for {run_date}")

            feature_group = utils.load_module(pipeline.module)
            df = self._execute(feature_group, run_date, pipeline)
            self._write(df, info_date, target)
            records = self._check_written(info_date, target)
            logger.info(f"Generated {records} records")
            if records == 0:
                raise RuntimeError("No records generated")
            else:
                return records
        return 0

    def _run_pipeline(self, pipeline: PipelineConfig):
        """
        Run single pipeline for all required dates

        :param pipeline: pipeline cfg
        :return: success bool
        """
        target = Table(
            schema_path=pipeline.target.target_schema,
            class_name=pipeline.module.python_class,
            partition=pipeline.target.target_partition_column,
        )
        logger.info(f"Loaded pipeline {pipeline.name}")

        selected_run_dates, selected_info_dates = self._select_run_dates(pipeline, target)

        # ----------- Checking dependencies available ----------
        for run_date, info_date in zip(selected_run_dates, selected_info_dates):
            run_start = datetime.datetime.now()
            try:
                records = self._run_one_date(pipeline, run_date, info_date, target)
                if records > 0:
                    status = "Success"
                    message = ""
                else:
                    status = "Failure"
                    message = self.tracker.last_error
                self.tracker.add(
                    Record(
                        job=pipeline.name,
                        target=target.get_table_path(),
                        date=info_date,
                        time=datetime.datetime.now() - run_start,
                        records=records,
                        status=status,
                        reason=message,
                    )
                )
            except Exception as error:
                logger.error(f"An exception occurred in pipeline {pipeline.name}")
                logger.exception(error)
                self.tracker.add(
                    Record(
                        job=pipeline.name,
                        target=target.get_table_path(),
                        date=info_date,
                        time=datetime.datetime.now() - run_start,
                        records=0,
                        status="Error",
                        reason="Exception",
                        exception=str(error),
                    )
                )
            except KeyboardInterrupt:
                logger.error(f"Pipeline {pipeline.name} interrupted")
                self.tracker.add(
                    Record(
                        job=pipeline.name,
                        target=target.get_table_path(),
                        date=info_date,
                        time=datetime.datetime.now() - run_start,
                        records=0,
                        status="Error",
                        reason="Interrupted by user",
                    )
                )
                raise KeyboardInterrupt

    def __call__(self):
        """Execute pipelines"""
        logger.info("Executing pipelines")
        try:
            if self.op:
                selected = [p for p in self.config.pipelines if p.name == self.op]
                if len(selected) < 1:
                    raise ValueError(f"Unknown operation selected: {self.op}")
                self._run_pipeline(selected[0])
            else:
                for pipeline in self.config.pipelines:
                    self._run_pipeline(pipeline)
        finally:
            print(self.tracker.records)
            self.tracker.report_by_mail()
            logger.info("Execution finished")



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["datasource", "job", "config_parser"]

import typing

import importlib_metadata
from loguru import logger

from rialto.common.utils import get_caller_module
from rialto.jobs.job_base import JobBase, JobMetadata
from rialto.jobs.module_register import ModuleRegister


def config_parser(cf_getter: typing.Callable) -> typing.Callable:
    """
    Config parser functions decorator.

    Registers a config parsing function into a rialto job prerequisite.
    You can then request the job via job function arguments.

    :param cf_getter:  dataset reader function
    :return: raw function, unchanged
    """
    ModuleRegister.register_callable(cf_getter)
    return cf_getter


def datasource(ds_getter: typing.Callable) -> typing.Callable:
    """
    Dataset reader functions decorator.

    Registers a data-reading function into a rialto job prerequisite.
    You can then request the job via job function arguments.

    :param ds_getter:  dataset reader function
    :return: raw reader function, unchanged
    """
    ModuleRegister.register_callable(ds_getter)
    return ds_getter


def _get_job_metadata(module: typing.Any) -> JobMetadata:
    try:
        package_name, _, _ = module.__name__.partition(".")
        dist_name = importlib_metadata.packages_distributions()[package_name][0]
        dist_version = importlib_metadata.version(dist_name)

        return JobMetadata(job_name=module.__name__, dist_name=dist_name, dist_version=dist_version)

    except Exception:
        logger.warning(f"Failed to get version of {module.__name__}! Will use N/A")
        return JobMetadata(job_name=module.__name__, dist_name="N/A", dist_version="N/A")


def _generate_rialto_job(
    callable: typing.Callable, module: object, class_name: str, metadata: JobMetadata, disable_version: bool
) -> typing.Type:
    generated_class = type(
        class_name,
        (JobBase,),
        {
            "get_custom_callable": lambda _: callable,
            "get_job_metadata": lambda _: metadata,
            "get_job_name": lambda _: class_name,
            "get_disable_version": lambda _: disable_version,
        },
    )

    generated_class.__module__ = module.__name__
    setattr(module, class_name, generated_class)

    logger.info(f"Registered {class_name} in {module}")
    return generated_class


def job(*args, custom_name=None, disable_version=False):
    """
    Rialto jobs decorator.

    Transforms a python function into a rialto transformation, which can be imported and ran by Rialto Runner.
    Is mainly used as @job and the function's name is used, and the outputs get automatic.
    To override this behavior, use @job(custom_name=XXX, disable_version=True).


    :param *args:  list of positional arguments. Empty in case custom_name or disable_version is specified.
    :param custom_name:  str for custom job name.
    :param disable_version:  bool for disabling automatically filling the VERSION column in the job's outputs.
    :return: One more job wrapper for run function (if custom name or version override specified).
             Otherwise, generates Rialto Transformation Type and returns it for in-module registration.
    """
    module = get_caller_module()
    metadata = _get_job_metadata(module)

    # Use case where it's just raw @f. Otherwise, we get [] here.
    if len(args) == 1 and callable(args[0]):
        f = args[0]
        return _generate_rialto_job(
            callable=f, module=module, class_name=f.__name__, metadata=metadata, disable_version=disable_version
        )

    # If custom args are specified, we need to return one more wrapper
    def inner_wrapper(f):
        # Setting default custom name, in case user only disables version
        class_name = f.__name__

        # User - Specified custom name
        if custom_name is not None:
            class_name = custom_name

        return _generate_rialto_job(
            callable=f, module=module, class_name=class_name, metadata=metadata, disable_version=disable_version
        )

    return inner_wrapper



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["MetadataManager"]

from typing import List

from delta.tables import DeltaTable

from rialto.metadata.data_classes.feature_metadata import FeatureMetadata
from rialto.metadata.data_classes.group_metadata import GroupMetadata


class MetadataManager:
    """Metadata storage i/o"""

    def __init__(self, session, schema_path: str = None):
        self.spark = session

        self.groups_path = f"{schema_path}.group_metadata"
        self.features_path = f"{schema_path}.feature_metadata"

        self.groups = None
        self.features = None

    def _load_metadata(self):
        if self.groups is None:
            self.groups = self.spark.read.table(self.groups_path)
        if self.features is None:
            self.features = self.spark.read.table(self.features_path)

    def _fetch_group_by_name(self, group_name: str) -> GroupMetadata:
        group = self.groups.filter(self.groups.group_name == group_name).collect()
        if not len(group):
            raise LookupError(f"Group {group_name} not found!")
        return GroupMetadata.from_spark(group[0])

    def _fetch_feature_by_name(self, feature_name: str, group_name: str) -> FeatureMetadata:
        feature = (
            self.features.filter(self.features.group_name == group_name)
            .filter(self.features.feature_name == feature_name)
            .collect()
        )
        if not len(feature):
            raise LookupError(f"Feature {feature_name} in group {group_name} not found!")
        return FeatureMetadata.from_spark(feature[0])

    def _fetch_features(self, group_name: str) -> List:
        return self.features.filter(self.features.group_name == group_name).collect()

    def _add_group(self, group_md: GroupMetadata) -> None:
        groups = DeltaTable.forName(self.spark, self.groups_path)
        df = self.spark.createDataFrame([group_md.to_tuple()], self.groups.schema)

        groups.alias("groups").merge(
            df.alias("updates"), "groups.group_name = updates.group_name "
        ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

    def _add_features(self, feature_md: List[FeatureMetadata], group_name: str) -> None:
        features = DeltaTable.forName(self.spark, self.features_path)
        feature_data = [md.to_tuple(group_name) for md in feature_md]
        df = self.spark.createDataFrame(feature_data, self.features.schema)

        features.alias("features").merge(
            df.alias("updates"),
            "features.feature_name = updates.feature_name and features.group_name = updates.group_name",
        ).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

    def update(
        self,
        group_md: GroupMetadata,
        features_md: List[FeatureMetadata],
    ):
        """
        Save or refresh information about generated features and their groups

        :param group_md: GroupMetadata object
        :param features_md: list of FeatureMetadata objects
        :return:
        """
        self._load_metadata()
        self._add_group(group_md)
        self._add_features(features_md, group_md.name)

    def get_feature(self, group_name: str, feature_name: str) -> FeatureMetadata:
        """
        Get metadata of one feature

        :param group_name: string name of feature group
        :param feature_name: string name of feature
        :return: FeatureMetadata object
        """
        self._load_metadata()
        group = self.get_group(group_name)
        feature = self._fetch_feature_by_name(feature_name, group_name)
        return feature.add_group(group)

    def get_group(self, group_name: str) -> GroupMetadata:
        """
        Get metadata of one feature group

        :param group_name: string name of feature group
        :return: GroupMetadata object
        """
        self._load_metadata()
        group = self._fetch_group_by_name(group_name)
        features = self._fetch_features(group_name)
        group.add_features([f.feature_name for f in features])
        return group



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["override_config"]

from typing import Dict, List, Tuple

from loguru import logger


def _split_index_key(key: str) -> Tuple[str, str]:
    name = key.split("[")[0]
    index = key.split("[")[1].replace("]", "")
    return name, index


def _find_first_match(config: List, index: str) -> int:
    index_key, index_value = index.split("=")
    return next(i for i, x in enumerate(config) if x.get(index_key) == index_value)


def _override(config, path, value) -> Dict:
    key = path[0]
    if "[" in key:
        name, index = _split_index_key(key)
        if name not in config:
            raise ValueError(f"Invalid key: {name}")
        if "=" in index:
            index = _find_first_match(config[name], index)
        else:
            index = int(index)
        if index >= 0 and index < len(config[name]):
            if len(path) == 1:
                config[name][index] = value
            else:
                config[name][index] = _override(config[name][index], path[1:], value)
        elif index == -1:
            if len(path) == 1:
                config[name].append(value)
            else:
                raise ValueError(f"Invalid index {index} for key {name} in path {path}")
        else:
            raise IndexError(f"Index {index} out of bounds for key {key}")
    else:
        if len(path) == 1:
            if key not in config:
                logger.warning(f"Adding new key: {key} with value {value}")
            config[key] = value
        else:
            if key not in config:
                raise ValueError(f"Invalid key: {key}")
            config[key] = _override(config[key], path[1:], value)
    return config


def override_config(config: Dict, overrides: Dict) -> Dict:
    """Override config with user input

    :param config: config dictionary
    :param overrides: dictionary of overrides
    :return: Overridden config
    """
    for path, value in overrides.items():
        logger.info(f"Applying override:\npath: {path}\nvalue: {value}")
        config = _override(config, path.split("."), value)

    return config



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["feature", "desc", "param", "depends"]

import typing
from copy import deepcopy
from functools import partial, wraps

from loguru import logger

from rialto.maker.containers import FeatureFunction, FeatureHolder
from rialto.metadata import ValueType


def decorator_with_args(decorator_to_enhance):
    """
    Wrap decorators to use args

    This function is supposed to be used as a decorator.
    It must decorate and other function, that is intended to be used as a decorator.
    It will allow any decorator to accept an arbitrary number of arguments
    """

    @wraps(decorator_to_enhance)
    def decorator_maker(*args, **kwargs):
        """
        Pass arguments to inner wrapper

        We create on the fly a decorator that accepts only a function
        but keeps the passed arguments from the maker.
        """

        def decorator_wrapper(func):
            """
            Pass arguments to the wrapped function

            We return the result of the original decorator
            the decorator must have this specific signature otherwise it won't work:
            """
            return decorator_to_enhance(func, *args, **kwargs)

        return decorator_wrapper

    return decorator_maker


@decorator_with_args
def feature(feature_functions: typing.Union[typing.Callable, FeatureHolder], feature_type: ValueType):
    """
    Wrap a feature definitions. Use as the outermost decorator!!!

    The purpose of this wrapper is to wrap function that have no other decorator inside FeatureHolder
    and define feature type
    :param feature_functions: FeatureHolder or pure function
    :param feature_type: FeatureType enum (numerical, ordinal, nominal)
    :return: FeatureHolder
    """
    logger.trace(f"Wrapping {feature_functions} as feature")

    def wrapper() -> FeatureHolder:
        if isinstance(feature_functions, FeatureHolder):
            for f in feature_functions:
                f.type = feature_type
            return feature_functions
        else:
            func_list = FeatureHolder()
            new_feature_f = FeatureFunction(feature_functions.__name__, feature_functions, feature_type)
            func_list.append(new_feature_f)
            return func_list

    return wrapper()


@decorator_with_args
def desc(feature_functions: typing.Union[typing.Callable, FeatureHolder], desc: str):
    """
    Wrap feature with string description, used in metadata

    :param feature_functions: FeatureHolder or pure function
    :param type: FeatureType enum (numerical, ordinal, nominal)
    :return: FeatureHolder
    """
    logger.trace(f"Wrapping {feature_functions} with description")

    def wrapper() -> FeatureHolder:
        if isinstance(feature_functions, FeatureHolder):
            for f in feature_functions:
                f.description = desc
            return feature_functions
        else:
            func_list = FeatureHolder()
            new_feature_f = FeatureFunction(feature_functions.__name__, feature_functions)
            new_feature_f.description = desc
            func_list.append(new_feature_f)
            return func_list

    return wrapper()


@decorator_with_args
def param(feature_functions: typing.Union[typing.Callable, FeatureHolder], parameter_name: str, values: typing.List):
    """
    Wrap a feature function with custom parameter, can be chained

    Creates a cartesian product of all previous parameters with new ones
    :param feature_functions: FeatureHolder or pure function
    :param parameter_name: string name of the parameter
    :param values: a list of parameter values
    :return: FeatureHolder
    """
    logger.trace(f"Parametrize \n\tfunc: {feature_functions}\n\tparam:{parameter_name}\n\tvalues{values}")

    def expander() -> FeatureHolder:
        func_list = FeatureHolder()
        if isinstance(feature_functions, FeatureHolder):
            for f in feature_functions:
                for value in values:
                    func_list.append(_copy_with_new_parameter(f, parameter_name, value))
        else:
            for value in values:
                func_list.append(_create_new_with_parameter(feature_functions, parameter_name, value))
        return func_list

    return expander()


def _copy_with_new_parameter(func: FeatureFunction, parameter: str, value: typing.Any) -> FeatureFunction:
    """
    Create a deep copy of FeatureFunction and store new parameter in it

    :param func: FeatureFunction
    :param parameter: string parameter name
    :param value: any parameter value
    :return: FeatureFunction
    """
    logger.trace(f"Extending \n\tfunction: {func} \n\tby parameter: {parameter} \n\twith value: {value}")
    new_feature_f = deepcopy(func)
    new_feature_f.callable = partial(func.callable, **{parameter: value})
    new_feature_f.parameters[parameter] = value
    return new_feature_f


def _create_new_with_parameter(func: typing.Callable, parameter: str, value: typing.Any) -> FeatureFunction:
    """
    Create a new FeatureFunction from passed callable and store new parameter in it

    :param func: callable object
    :param parameter: string parameter name
    :param value: any parameter value
    :return: FeatureFunction
    """
    logger.trace(f"Creating \n\tfunction: {func} \n\twith parameter: {parameter} \n\twith value: {value}")
    new_feature_f = FeatureFunction(func.__name__, partial(func, **{parameter: value}))
    new_feature_f.parameters[parameter] = value
    return new_feature_f


@decorator_with_args
def depends(feature_functions: typing.Union[typing.Callable, FeatureHolder], dependency: str):
    """
    Specify dependency features

    The purpose of this wrapper is to define inter-feature dependency and order the execution accordingly
    :param feature_functions: FeatureHolder or pure function
    :param dependency: str name of required function
    :return: FeatureHolder
    """
    logger.trace(f"Wrapping {feature_functions} with dependency {dependency}")

    def wrapper() -> FeatureHolder:
        if isinstance(feature_functions, FeatureHolder):
            for f in feature_functions:
                f.dependencies.append(dependency)
            return feature_functions
        else:
            func_list = FeatureHolder()
            new_feature_f = FeatureFunction(feature_functions.__name__, feature_functions)
            new_feature_f.dependencies.append(dependency)
            func_list.append(new_feature_f)
            return func_list

    return wrapper()



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["load_yaml", "cast_decimals_to_floats", "get_caller_module"]

import inspect
import os
from typing import Any, List

import pyspark.sql.functions as F
import yaml
from pyspark.sql import DataFrame
from pyspark.sql.types import FloatType

from rialto.common.env_yaml import EnvLoader


def load_yaml(path: str) -> Any:
    """
    YAML loader

    :param path: file path
    :return: Parsed yaml
    """
    if not os.path.isfile(path):
        raise FileNotFoundError(f"Can't find {path}.")

    with open(path, "r") as stream:
        return yaml.load(stream, EnvLoader)


def cast_decimals_to_floats(df: DataFrame) -> DataFrame:
    """
    Find all decimal types in the table and cast them to floats. Fixes errors in .toPandas() conversions.

    :param df: input df
    :return: pyspark DataFrame with fixed types
    """
    decimal_cols = [col_name for col_name, data_type in df.dtypes if "decimal" in data_type]
    for c in decimal_cols:
        df = df.withColumn(c, F.col(c).cast(FloatType()))

    return df


def get_caller_module() -> Any:
    """
    Ged module containing the function which is calling your function.

    Inspects the call stack, where:
    0th entry is this function
    1st entry is the function which needs to know who called it
    2nd entry is the calling function

    Therefore, we'll return a module which contains the function at the 2nd place on the stack.

    :return: Python Module containing the calling function.
    """

    stack = inspect.stack()
    last_stack = stack[2]
    return inspect.getmodule(last_stack[0])



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["Tracker"]

from datetime import datetime

from pyspark.sql import SparkSession

from rialto.runner.config_loader import MailConfig
from rialto.runner.reporting.bookkeeper import BookKeeper
from rialto.runner.reporting.mailer import HTMLMessage, Mailer
from rialto.runner.reporting.record import Record


class Tracker:
    """Collect information about runs and sent them out via email"""

    def __init__(self, mail_cfg: MailConfig, bookkeeping: str = None, spark: SparkSession = None):
        self.records = []
        self.last_error = None
        self.pipeline_start = datetime.now()
        self.exceptions = []
        self.mail_cfg = mail_cfg
        self.bookkeeper = None

        if bookkeeping:
            self.bookkeeper = BookKeeper(table=bookkeeping, spark=spark)

    def add(self, record: Record) -> None:
        """Add record for one run"""
        self.records.append(record)
        if self.bookkeeper:
            self.bookkeeper.add(record)

    def report_by_mail(self):
        """Create and send html report"""
        if self.mail_cfg is None:
            return
        if len(self.records) or self.mail_cfg.sent_empty:
            report = HTMLMessage.make_report(self.pipeline_start, self.records)
            for receiver in self.mail_cfg.to:
                message = Mailer.create_message(
                    subject=self.mail_cfg.subject, sender=self.mail_cfg.sender, receiver=receiver, body=report
                )
                Mailer.send_mail(self.mail_cfg.smtp, message)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import smtplib
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List

from rialto.runner.reporting.record import Record

__all__ = ["Mailer"]


class HTMLMessage:
    bck_colors = ["#00ded6", "#acfcfa"]
    borderless_table = 'role="presentation" style="border:0;border-spacing:0;"'
    bordered_table = (
        'role="presentation" style="background-repeat:no-repeat; margin:0;" cellpadding="1" cellspacing="1" border="1""'
    )

    @staticmethod
    def _get_status_color(status: str):
        if status == "Success":
            return "#398f00"
        elif status == "Error":
            return "#ff0000"
        else:
            return "#ff8800"

    @staticmethod
    def _make_rows(rows):
        html = ""
        data_options = 'align="center"'
        for row, i in zip(rows, range(len(rows))):
            r = f"""
                    <tr style="height:40px; margin:0; background-color:{HTMLMessage.bck_colors[i % 2]}">
                        <td {data_options}>{row.job}</td>
                        <td {data_options}>{row.target.split('.')[0]}.<br>
                                            {row.target.split('.')[1]}.<br>
                                            {row.target.split('.')[2]}
                        </td>
                        <td {data_options}>{row.date}</td>
                        <td {data_options}>{str(row.time).split(".")[0]}</td>
                        <td {data_options}>{f'{row.records:,}'}</td>
                        <td {data_options} style="color: {HTMLMessage._get_status_color(row.status)};">
                            <b>{row.status}</b>
                        </td>
                        <td {data_options}>{row.reason}</td>
                    </tr>
                """
            html += r
        return html

    @staticmethod
    def _make_overview_header():
        return """
            <tr style="height:40px; margin:2px; background-color: #286dd4;">
                <th>Job</th>
                <th>Target</th>
                <th>Date</th>
                <th>Time elapsed</th>
                <th>Rows created</th>
                <th>Status</th>
                <th>Reason</th>
            </tr>
        """

    @staticmethod
    def _make_header(start: datetime):
        return f"""
            <div align="center">
                <table {HTMLMessage.borderless_table}>
                    <tr>
                        <td align="center"><h3>This is a Rialto report<h3></td>
                    </tr>
                    <tr>
                        <td align="center">
                            Jobs started <b>{str(start).split('.')[0]}</b>
                        </td>
                    </tr>
                </table>
            </div>
        """

    @staticmethod
    def _make_overview(records: List[Record]):
        return f"""
            <table {HTMLMessage.borderless_table}>
                <tr>
                    <td><h3>Overview</h3></td>
                </tr>
            </table>
            <table {HTMLMessage.bordered_table}>
                {HTMLMessage._make_overview_header()}
                {HTMLMessage._make_rows(records)}
            </table>
        """

    @staticmethod
    def _head():
        return """
            <head>
                <meta charset="utf-8">
                <meta name="viewport" content="width=device-width,initial-scale=1">
                <meta name="x-apple-disable-message-reformatting">
                <!--[if !mso]><!-->
                    <meta http-equiv="X-UA-Compatible" content="IE=edge" />
                <!--<![endif]-->
                <!--[if mso]>
                    <noscript>
                        <xml>
                            <o:OfficeDocumentSettings>
                            <o:PixelsPerInch>96</o:PixelsPerInch>
                            </o:OfficeDocumentSettings>
                        </xml>
                    </noscript>
                <![endif]-->
            </head>
            <style>
                .foldingcheckbox { float: left; }
                .foldingcheckbox:not(:checked) + * { display: none }
            </style>
        """

    @staticmethod
    def _body_open():
        return """
            <body style="margin:0;padding:0;word-spacing:normal">
                <div role="article" aria-roledescription="email" lang="en"
                        style="-webkit-text-size-adjust:100%;-ms-text-size-adjust:100%;">
                    <div class="outer" style="width:100%;">
                """

    @staticmethod
    def _body_close():
        return """
                    </div>
                </div>
            </body>
        """

    @staticmethod
    def _make_exceptions(records: List[Record]):
        html = ""
        for record, i in zip(records, range(len(records))):
            if record.exception is not None:
                r = f"""
                    <table {HTMLMessage.bordered_table}>
                        <tr >
                            <td bgcolor={HTMLMessage.bck_colors[0]}>{record.job}</td>
                            <td bgcolor={HTMLMessage.bck_colors[1]}>{record.date}</td>
                        </tr>
                    </table>

                    <input class="foldingcheckbox" type="checkbox">Expand</input>
                    <div>
                        <table {HTMLMessage.borderless_table}>
                            <tr>
                                <td colspan="2">{record.exception}</td>
                            </tr>
                        </table>
                    </div>
                """
                html += r
        return html

    @staticmethod
    def _make_insights(records: List[Record]):
        return f"""
            <table {HTMLMessage.borderless_table}>
                <tr>
                    <td><h3>Exceptions<h3></td>
                </tr>
            </table>
            {HTMLMessage._make_exceptions(records)}
        """

    @staticmethod
    def make_report(start: datetime, records: List[Record]) -> str:
        """Create html email report"""
        html = [
            """<!DOCTYPE html>
            <html lang="en" xmlns="https://www.w3.org/1999/xhtml" xmlns:o="urn:schemas-microsoft-com:office:office">""",
            HTMLMessage._head(),
            HTMLMessage._body_open(),
            HTMLMessage._make_header(start),
            HTMLMessage._make_overview(records),
            HTMLMessage._make_insights(records),
            HTMLMessage._body_close(),
        ]
        return "\n".join(html)


class Mailer:
    """Send email reports"""

    @staticmethod
    def create_message(subject: str, sender: str, receiver: str, body: str) -> MIMEMultipart:
        """
        Create email message

        :param subject: email subject
        :param sender: email sender
        :param receiver: email receiver
        :param body: email body
        """
        msg = MIMEMultipart()
        msg["Subject"] = subject
        msg["From"] = sender
        msg["To"] = receiver
        body = MIMEText(body, "html")
        msg.attach(body)
        return msg

    @staticmethod
    def send_mail(smtp: str, message: MIMEMultipart):
        """
        Send email

        :param smtp: smtp server
        :param message: email message
        """
        s = smtplib.SMTP(host=smtp, port=25)
        s.sendmail(from_addr=message["From"], to_addrs=message["To"], msg=message.as_string())
        s.quit()



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["PysparkFeatureLoader"]

from collections import namedtuple
from datetime import date
from typing import Dict, List, Union

from pyspark.sql import DataFrame, SparkSession

from rialto.common import TableReader
from rialto.common.utils import cast_decimals_to_floats
from rialto.loader.config_loader import FeatureConfig, GroupConfig, get_feature_config
from rialto.loader.interfaces import FeatureLoaderInterface
from rialto.metadata.metadata_manager import (
    FeatureMetadata,
    GroupMetadata,
    MetadataManager,
)


class PysparkFeatureLoader(FeatureLoaderInterface):
    """Implementation of feature loader for pyspark environment"""

    def __init__(
        self,
        spark: SparkSession,
        feature_schema: Union[str, List[str]],
        metadata_schema: str,
        date_column: str = "INFORMATION_DATE",
    ):
        """
        Init

        :param spark: spark session
        :param data_loader: data loader
        :param metadata_schema: schema location of metadata tables
        """
        super().__init__()
        self.spark = spark
        self.reader = TableReader(spark)

        if isinstance(feature_schema, str):
            feature_schema = [feature_schema]

        self.feature_schemas = feature_schema
        self.date_col = date_column
        self.metadata = MetadataManager(spark, metadata_schema)

    KeyMap = namedtuple("KeyMap", ["df", "key"])

    def read_group(self, group: str, information_date: date) -> DataFrame:
        """
        Read a feature group by getting the latest partition by date

        :param group: group name
        :param information_date: partition date
        :return: dataframe
        """
        selected = []

        for schema in self.feature_schemas:
            tables = self.spark.catalog.listTables(schema)
            if any(table.name == group for table in tables):
                selected.append(schema)
        if len(selected) > 1:
            raise ValueError(f"Multiple feature schemas contain table {group}: {selected}.")
        elif len(selected) == 0:
            raise ValueError(f"Feature group {group} not found in schemas {self.feature_schemas}.")
        else:
            return self.reader.get_latest(
                f"{selected[0]}.{group}",
                date_until=information_date,
                date_column=self.date_col,
                uppercase_columns=True,
            )

    def get_feature(self, group_name: str, feature_name: str, information_date: date) -> DataFrame:
        """
        Get single feature

        :param group_name: feature group name
        :param feature_name: feature name
        :param information_date: selected date
        :return: A dataframe containing feature group key and selected feature
        """
        print("This function is untested, use with caution!")
        key = self.get_group_metadata(group_name).key
        return self.read_group(self.get_group_fs_name(group_name), information_date).select(*key, feature_name)

    def get_feature_metadata(self, group_name: str, feature_name: str) -> FeatureMetadata:
        """
        Get single features metadata

        :param group_name: feature group name
        :param feature_name: feature name
        :return: metadata dictionary
        """
        return self.metadata.get_feature(group_name, feature_name)

    def get_group(self, group_name: str, information_date: date) -> DataFrame:
        """
        Get feature group

        :param group_name: feature group name
        :param information_date: selected date
        :return: A dataframe containing feature group key
        """
        print("This function is untested, use with caution!")
        return self.read_group(self.get_group_fs_name(group_name), information_date)

    def get_group_metadata(self, group_name: str) -> GroupMetadata:
        """
        Get feature groups metadata

        :param group_name: feature group name
        :return: metadata dictionary
        """
        return self.metadata.get_group(group_name)

    def get_group_fs_name(self, group_name: str) -> str:
        """
        Return groups file system name

        If the name is already lowercase, assume it's a filesystem name.
        Else assume it's a class name and search for fs name in metadata.
        :param group_name: Group name
        :return: group filesystem name
        """
        if group_name.islower():
            return group_name
        return self.metadata.get_group(group_name).fs_name

    def _are_all_keys_in(self, keys: List[str], columns: List[str]) -> bool:
        """
        Check if all presented keys are in presented list of columns

        :param keys: list of keys
        :param columns: list of columns
        :return: True/False
        """
        for key in keys:
            if key not in columns:
                return False
        return True

    def _add_prefix(self, df: DataFrame, prefix: str, key: List[str]) -> DataFrame:
        """
        Prefixes all column names except for key

        :param df: dataframe
        :param prefix: prefix
        :param key: list of keys
        :return: renamed dataframe
        """
        for col in df.columns:
            if col not in key:
                df = df.withColumnRenamed(col, f"{prefix}_{col}")
        return df

    def _get_keymaps(self, config: FeatureConfig, information_date: date) -> List[KeyMap]:
        """
        Read all key mapping tables specified by configuration

        :param config: configuration object
        :param information_date: date
        :return: List of tuples of loaded dataframes and their keys
        """
        key_maps = []
        for mapping in config.maps:
            df = self.read_group(self.get_group_fs_name(mapping), information_date).drop("INFORMATION_DATE")
            key = self.metadata.get_group(mapping).key
            key_maps.append(PysparkFeatureLoader.KeyMap(df, key))
        return key_maps

    def _join_keymaps(self, base: DataFrame, key_maps: List[KeyMap]) -> DataFrame:
        if len(key_maps):
            for mapping in key_maps:
                if self._are_all_keys_in(mapping.key, base.columns):
                    base = base.join(mapping.df, mapping.key, "inner")
                    key_maps.remove(mapping)
                    return self._join_keymaps(base, key_maps)
            raise KeyError(f"None of {[x.key for x in key_maps]} can be joined onto {base.columns}")
        else:
            return base

    def _add_feature_group(self, base: DataFrame, df: DataFrame, group_cfg: GroupConfig) -> DataFrame:
        group_key = self.metadata.get_group(group_cfg.group).key
        df = df.select(group_cfg.features + group_key)
        df = self._add_prefix(df, group_cfg.prefix, group_key)
        return base.join(df, group_key, "left")

    def get_features_from_cfg(self, path: str, information_date: date) -> DataFrame:
        """Get multiple features across many groups, fetches latest features in relation to the provided date

        :param path: configuration location
        :param information_date: date for extraction
        """
        config = get_feature_config(path)
        # 1 select keys from base
        base = self.read_group(self.get_group_fs_name(config.base.group), information_date).select(config.base.keys)
        # 2 join maps onto base (resolve keys)
        if config.maps:
            key_maps = self._get_keymaps(config, information_date)
            base = self._join_keymaps(base, key_maps)

        # 3 read, select and join other tables
        for group_cfg in config.selection:
            df = self.read_group(self.get_group_fs_name(group_cfg.group), information_date)
            base = self._add_feature_group(base, df, group_cfg)

        # 4 fix dtypes for pandas conversion
        base = cast_decimals_to_floats(base)

        return base

    def get_metadata_from_cfg(self, path: str) -> Dict[str, FeatureMetadata]:
        """
        Get multiple features metadata from config

        :param path: configuration path
        :return: dictionary feature_name : FeatureMetadata
        """
        result = {}
        config = get_feature_config(path)

        for group_cfg in config.selection:
            for feature in group_cfg.features:
                feature_metadata = self.get_feature_metadata(group_cfg.group, feature)
                feature_name = f"{group_cfg.prefix}_{feature}"
                result[feature_name] = feature_metadata

        return result



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pyspark.sql import DataFrame, SparkSession

from rialto.runner.reporting.record import Record


class BookKeeper:
    """Class to store and update records of runs in a table in the spark catalog."""

    def __init__(self, table: str, spark: SparkSession):
        self.spark = spark
        self.table = table

    def _write(self, df: DataFrame) -> None:
        if self.spark.catalog.tableExists(self.table):
            df.write.mode("append").saveAsTable(self.table)
        else:
            df.write.mode("overwrite").saveAsTable(self.table)

    def add(self, record: Record) -> None:
        """
        Add a record to the table.

        :param record: Record to add to the table.
        """
        new = self.spark.createDataFrame([record.to_spark_row()], record.get_schema())

        self._write(new)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["FeatureMetadata"]

from dataclasses import dataclass
from typing import Tuple

from pyspark.sql import Row
from typing_extensions import Self

from rialto.metadata.data_classes.group_metadata import GroupMetadata
from rialto.metadata.enums import ValueType


@dataclass
class FeatureMetadata:
    """A dataclass to hold all information about a feature"""

    value_type: ValueType
    name: str
    description: str
    group: GroupMetadata = None

    def __repr__(self) -> str:
        """Serialize object to string"""
        return (
            "FeatureMetadata("
            f"name={self.name!r}, value_type={self.value_type!r}, "
            f"description={self.description!r}, group={self.group!r}, "
            ")"
        )

    def to_tuple(self, group_name: str) -> Tuple:
        """
        Serialize to tuple

        :param group_name: Feature group name
        :return: tuple with feature information
        """
        return (self.name, self.value_type.value, self.description, group_name)

    def add_group(self, group: GroupMetadata) -> Self:
        """
        Add group information to metadata

        :param group: Group name
        :return: self
        """
        self.group = group
        return self

    @classmethod
    def from_spark(cls, record: Row) -> Self:
        """
        Create new instance from spark row

        :param record: spark row
        :return: new instance
        """
        return FeatureMetadata(
            value_type=ValueType[record.feature_type],
            name=record.feature_name,
            description=record.feature_description,
        )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["FeatureLoaderInterface"]

import abc
from datetime import date
from typing import Dict


class FeatureLoaderInterface(metaclass=abc.ABCMeta):
    """
    A definition of feature loading interface

    Provides functionality to read features, feature groups and selections of features according to configs.
    Also provides an interface to access metadata of said features.
    """

    @abc.abstractmethod
    def get_feature(self, group_name: str, feature_name: str, information_date: date):
        """Get single feature"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_feature_metadata(self, group_name: str, feature_name: str) -> Dict:
        """Get single feature's metadata"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_group(self, group_name: str, information_date: date):
        """Get feature group"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_group_metadata(self, group_name: str) -> Dict:
        """Get feature group's metadata"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_features_from_cfg(self, path: str, information_date: date):
        """Get features from multiple groups as defined by configuration"""
        raise NotImplementedError

    @abc.abstractmethod
    def get_metadata_from_cfg(self, path: str) -> Dict:
        """Get metadata from multiple groups as defined by configuration"""
        raise NotImplementedError



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["disable_job_decorators"]

import importlib
import typing
from contextlib import contextmanager
from unittest.mock import MagicMock, create_autospec, patch

from rialto.jobs.job_base import JobBase
from rialto.jobs.module_register import ModuleRegister
from rialto.jobs.resolver import Resolver, ResolverException


def _passthrough_decorator(*args, **kwargs) -> typing.Callable:
    if len(args) == 0:
        return _passthrough_decorator
    else:
        return args[0]


@contextmanager
def _disable_job_decorators() -> None:
    patches = [
        patch("rialto.jobs.datasource", _passthrough_decorator),
        patch("rialto.jobs.decorators.datasource", _passthrough_decorator),
        patch("rialto.jobs.config_parser", _passthrough_decorator),
        patch("rialto.jobs.decorators.config_parser", _passthrough_decorator),
        patch("rialto.jobs.job", _passthrough_decorator),
        patch("rialto.jobs.decorators.job", _passthrough_decorator),
    ]

    for i in patches:
        i.start()

    yield

    for i in patches:
        i.stop()


@contextmanager
def disable_job_decorators(module) -> None:
    """
    Disables job decorators in a python module. Useful for testing your rialto jobs and datasources.

    :param module: python module with the decorated functions.
    :return: None
    """
    with _disable_job_decorators():
        ModuleRegister.remove_module(module)
        importlib.reload(module)
        yield

    ModuleRegister.remove_module(module)
    importlib.reload(module)


def resolver_resolves(spark, job: JobBase) -> bool:
    """
    Checker method for your dependency resolution.

    If your job's dependencies are all defined and resolvable, returns true.
    Otherwise, throws an exception.

    :param spark: SparkSession object.
    :param job: Job to try and resolve.

    :return: bool, True if job can be resolved
    """
    call_stack = []
    original_resolve_method = Resolver.resolve

    def stack_watching_resolver_resolve(self, callable):
        # Check for cycles
        if callable in call_stack:
            raise ResolverException(f"Circular Dependence in {callable.__name__}!")

        # Append to call stack
        call_stack.append(callable)

        # Create fake method
        fake_method = create_autospec(callable)
        fake_method.__module__ = callable.__module__

        # Resolve fake method
        result = original_resolve_method(self, fake_method)

        # Remove from call stack
        call_stack.remove(callable)

        return result

    with patch("rialto.jobs.job_base.Resolver.resolve", stack_watching_resolver_resolve):
        with patch("rialto.jobs.job_base.JobBase._add_job_version", lambda _, x: x):
            job().run(
                reader=MagicMock(),
                run_date=MagicMock(),
                spark=spark,
                config=MagicMock(),
                metadata_manager=MagicMock(),
                feature_loader=MagicMock(),
            )
        return True



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = ["ModuleRegister", "register_dependency_module", "register_dependency_callable"]

from rialto.common.utils import get_caller_module


class ModuleRegisterException(Exception):
    """Module Register Exception - Usually, means a clash in the dependencies"""

    pass


class ModuleRegister:
    """
    Module register. Class which is used by @datasource and @config_parser decorators to register callables / getters.

    Resolver, when searching for a getter for f() defined in module M, uses find_callable("f", "M").
    """

    _storage = {}
    _dependency_tree = {}

    @classmethod
    def add_callable_to_module(cls, callable, parent_name):
        """
        Add a callable to the specified module's storage.

        :param callable: The callable to be added.
        :param parent_name: The name of the module to which the callable is added.
        """
        module_callables = cls._storage.get(parent_name, [])
        module_callables.append(callable)

        cls._storage[parent_name] = module_callables

    @classmethod
    def register_callable(cls, callable):
        """
        Register a callable by adding it to the module's storage.

        :param callable: The callable to be registered.
        """
        callable_module = callable.__module__
        cls.add_callable_to_module(callable, callable_module)

    @classmethod
    def remove_module(cls, module):
        """
        Remove a module from the storage.

        :param module: The module to be removed.
        """
        cls._storage.pop(module.__name__, None)
        cls._dependency_tree.pop(module.__name__, None)

    @classmethod
    def register_dependency(cls, module, parent_name):
        """
        Register a module as a dependency of the caller module.

        :param module: The module to be registered as a dependency.
        :param parent_name: The module that is registering the dependency.

        """
        module_dep_tree = cls._dependency_tree.get(parent_name, [])
        module_dep_tree.append(module)

        cls._dependency_tree[parent_name] = module_dep_tree

    @classmethod
    def find_callable(cls, callable_name, module_name):
        """
        Find a callable by its name in the specified module and its dependencies.

        :param callable_name: The name of the callable to find.
        :param module_name: The name of the module to search in.
        :return: The found callable or None if not found.
        """
        found_functions = []

        # Loop through this module, and its dependencies
        searched_modules = [module_name] + cls._dependency_tree.get(module_name, [])
        for module in searched_modules:
            # Loop through all functions registered in the module
            for func in cls._storage.get(module, []):
                if func.__name__ == callable_name:
                    found_functions.append(func)

        if len(found_functions) == 0:
            return None

        if len(found_functions) > 1:
            raise ModuleRegisterException(f"Multiple functions with the same name {callable_name} found !")

        else:
            return found_functions[0]


def register_dependency_module(module):
    """
    Register a module as a dependency of the caller module.

    :param module: The module to be registered as a dependency.
    """
    caller_module = get_caller_module().__name__
    ModuleRegister.register_dependency(module.__name__, caller_module)


def register_dependency_callable(callable):
    """
    Register a callable as a dependency of the caller module.

    Note that the function will be added to the module's list of available dependencies.

    :param callable: The callable to be registered as a dependency.
    """
    caller_module = get_caller_module().__name__
    ModuleRegister.add_callable_to_module(callable, caller_module)
